#!/usr/bin/env python3
#
# Author:  Carlos E. Budde
# Date:    06.10.2019
# License: GPLv3
#

import re
from typing import List
from os import path
from io import TextIOWrapper
from sys import argv, exit
from mmap import mmap, ACCESS_READ
from numpy import inf, abs, mean, median, std, array

if len(argv) != 2:
	print("[ERROR] Must call with path to file with summarised results")
	exit(1)
elif not path.isfile(argv[1]):
	print("[ERROR] File \""+argv[1]+"\" not found")
	exit(1)


########################################
##
##  CONFIGURATION
##
# Regex: whatever is considered a number: 1, 2.34, 1.23e-4...
ARBITRARY_NUM = "inf|[0-9]+(?:\.[0-9]+(?:(?:e|E)(?:\-|\+)[0-9]*)*)*"
# Regex: heading of experiment-run: "FLAT", "RES_ES", "FE_HYB8"...
RUN_LINE_REGEX = br"^[A-Z]{2,}[\.0-9]*(?:_)*(?:[A-Z]*[0-9]*)*\n"
# Regex: heading of model tested: "Model: tandem queue", "Model: 2tandem_NM"...
MODEL_LINE_REGEX = bytes("Model:.*","utf8")


########################################
##
##  SCRIPT
##

def trimb(bstr: bytes) -> str:
	return str(bstr)[2:-3]

def eat_spaces(fc: mmap) -> mmap:
	pos = fc.tell()
	line = fc.readline()
	while line.isspace():
		pos = fc.tell()
		line = fc.readline()
	fc.seek(pos)
	return fc

def find_numbers(fc: mmap) -> List[str]:
	"""
	Uses empty lines as delimiters: first skips starting empty lines,
	and stops on the first empty line encountered after the first number.
	fc is left at the position after reading all numbers.
	"""
	numbers = ""
	fc = eat_spaces(fc)
	line = fc.readline()
	while not line.isspace():
		numbers += " " + trimb(line)
		line = fc.readline()
	return re.findall(ARBITRARY_NUM, numbers)

def robust_stats(values: List[float]) -> (float,float,int):
	"""
	Filter values passed, to keep those strictly between 0.0 and 1.0,
	and then also discard those which are too far away from the median.
	From the resulting "robust subset" return the mean, stdev, and size.
	Robust mean: https://stackoverflow.com/a/16562028
	"""
	data = array([ float(v) for v in values if 0.0 < float(v) < 1.0 ])
	num_valid_datapoints = len(data)
	if num_valid_datapoints < 1:
		return 0,inf,0
	elif num_valid_datapoints < 2:
		return data[0],0,1
	dev = abs(data - median(data))
	mdev = median(dev)
	s = dev/mdev if mdev != 0.0 else 0.0
	m = 2  # outlier level: the greater, the fewer values filtered
	filtered_values = [data[s<m]]
	if len(filtered_values) > 0:
		rob_mean = mean(filtered_values)
		rob_stdev = std(filtered_values)
	else:
		rob_mean = 0
		rob_stdev = inf
	return rob_mean,rob_stdev,num_valid_datapoints

def read_experiment(fc: mmap, run: str, df: TextIOWrapper) -> List[float]:
	"""
	Read a whole experiment run, e.g. FLAT with all models.
	Write (precision) data read in df, formatted for gnuplot CI plotting.
	"""
	# Find run line in fc file
	fc.seek(0)
	fc.seek(fc.find(bytes(run,"utf8")))
	line = fc.readline()
	fc = eat_spaces(fc)
	line = fc.readline()
	df.write("# x avg      stdev    N\n")
	cnt = 0
	pmeans = []
	# For each model ...
	while type(None) != type(re.match(MODEL_LINE_REGEX,line)):
		# ... identify the model name of these precision values ...
		MODEL = re.findall(".*:\s*(\w+)", trimb(line))[0].replace("_","-")
		# ... and dump the mean,stdev,count of (a robust subset of) the values 
		stats = [ "{:.2e}".format(v) for v in robust_stats(find_numbers(fc)) ]
		stats[-1] = str(int(float(stats[-1])))
		pmeans += [float(stats[0])]
		df.write(" ".join([str(cnt),MODEL] + stats))
		df.write("\n")
		cnt += 1
		fc = eat_spaces(fc)
		line = fc.readline()
	return pmeans

# Invoke on file
PATH = argv[1]
print("Processing file " + PATH)
with open(PATH) as f, mmap(f.fileno(), 0, access=ACCESS_READ) as fc:
	# Identify experiments ran
	RUNS = set(re.findall(RUN_LINE_REGEX, fc, re.MULTILINE))
	RUNS = [ str(r)[2:-3] for r in RUNS ]
	RUNS.sort()
	FNAME = path.basename(PATH)
	DIR = path.dirname(PATH) if len(path.dirname(PATH))>0 else "."
	PREFIX = DIR + "/" + FNAME[ FNAME.find("_")+1 : FNAME.find(".") ]
	all_pmeans = []
	# Build a results file for each experiment run
	for run in RUNS:
		DATA_FNAME = PREFIX + "_" + run + ".dat"
		dfile = open(DATA_FNAME, "w")
		print("Reading results of experiment " + run, end="...")
		all_pmeans += [ read_experiment(fc, run, dfile) ]
		dfile.close()
		print("done: " + DATA_FNAME)
	fc.close()
	# Show in CLI the (mean) precisions read for all difficulties
	prec_ref_vals = []
	for i in range(len(all_pmeans[0])):
		pmeans_i = [ pmeans[i] for pmeans in all_pmeans ]
		prec_ref_vals += [ list(robust_stats(pmeans_i))[0:2] ]
	print("Created "+str(len(RUNS))+" data files with "
		 + str(len(prec_ref_vals)) +" models each.\n"
		 + "Precision- mean,stdev -of the 95% CIs built for each model:")
	for m,s in prec_ref_vals:
		print(" - " + "{:1.2e}".format(m) + "," + "{:1.2e}".format(s))

