#!/usr/bin/env python3

import re
import os
import sys
import signal
import subprocess
import datetime
from timeit import default_timer as timer


# Small quick run?
TEST_RUN = True


# Config experiments
TIMEOUT = 900  # in seconds
REPETITIONS = 3
RESULTS_FOLDER = "results"

# Paths to fig and modes binaries
FIG_BIN = "/home/tacas20ae/fig/fig"
MODEST_BIN = "/home/tacas20ae/Modest/modest"

for BIN in [FIG_BIN, MODEST_BIN]:
	if not os.path.isfile(BIN):
		print("Error: binary file "+str(BIN)+" not found")
	elif not os.access(BIN, os.X_OK):
		print("Error: binary file "+str(BIN)+" is not executable")

RESULTS_NUM = 0
while os.path.isdir(RESULTS_FOLDER + "_" + str(RESULTS_NUM)):
	RESULTS_NUM += 1
RESULTS_FOLDER += "_" + str(RESULTS_NUM) + "/"
os.mkdir(RESULTS_FOLDER)


model_instances = [
	("queues_with_breakdowns", "40"),
	("queues_with_breakdowns", "80"),
	("queues_with_breakdowns", "120"),
	("open_closed_queue", "20"),
	("open_closed_queue", "30"),
	("open_closed_queue", "40"),
	("tandem_queue", "8"),
	("tandem_queue", "10"),
	("tandem_queue", "12"),
	("database", "2-50"),
	("database", "3-50"),
	("database", "2-500"),
	("database", "3-500"),
	("oilpipes", "2-exp"),
	("oilpipes", "3-exp"),
	("oilpipes", "4-exp"),
	("oilpipes", "2-ray"),
	("oilpipes", "3-ray"),
	("oilpipes", "4-ray"),
]

tools = [
	("modes.restart.smc",
		[MODEST_BIN,
			"simulate",
			"models/{0}.{1}.jani",
			"-D",
			"--unsafe",
			"-J", "1",
			"-N", "2000000000",
			"--timeout", "{2}",
			"--rare", "Restart",
			"--levels", "SequentialMonteCarlo",
			"--max-run-length", "0"],
		re.compile(r"Confidence interval: *\[(.+), (.+)\]$", re.MULTILINE),
		re.compile(r"Runs used: *([0-9]+)$", re.MULTILINE)),
	("modes.restart.es",
		[MODEST_BIN,
			"simulate",
			"models/{0}.{1}.jani",
			"-D",
			"--unsafe",
			"-J", "1",
			"-N", "2000000000",
			"--timeout", "{2}",
			"--rare", "Restart",
			"--levels", "ExpectedSuccess",
			"--max-run-length", "0"],
		re.compile(r"Confidence interval: *\[(.+), (.+)\]$", re.MULTILINE),
		re.compile(r"Runs used: *([0-9]+)$", re.MULTILINE)),
	("modes.fixed.smc",
		[MODEST_BIN,
			"modes",
			"models/{0}.{1}.jani",
			"-D",
			"--unsafe",
			"-J", "1",
			"-N", "2000000000",
			"--timeout", "{2}",
			"--rare", "FixedEffort",
			"--levels", "SequentialMonteCarlo",
			"--max-run-length", "0"],
		re.compile(r"Confidence interval: *\[(.+), (.+)\]$", re.MULTILINE),
		re.compile(r"Runs used: *([0-9]+)$", re.MULTILINE)),
	("fig.restart.smc",
		[FIG_BIN,
			"models/{0}.{1}.sa",
			"--acomp",
			"--stop-time", "{2}",
			"--timeout", "{2}",
			"-e", "restart",
			"-t", "hyb"],
		re.compile(r"95% confidence\n.*\n.*interval: \[ (.+), (.+)\]$", re.MULTILINE),
		re.compile(r"([0-9]+) samples", re.MULTILINE)),
	("fig.restart.es",
		[FIG_BIN,
			"models/{0}.{1}.sa",
			"--acomp",
			"--stop-time", "{2}",
			"--timeout", "{2}",
			"-e", "restart",
			"-t", "es"],
		re.compile(r"95% confidence\n.*\n.*interval: \[ (.+), (.+)\]$", re.MULTILINE),
		re.compile(r"([0-9]+) samples", re.MULTILINE)),
	("fig.fixed.smc",
		[FIG_BIN,
			"models/{0}.{1}.sa",
			"--acomp",
			"--stop-time", "{2}",
			"--timeout", "{2}",
			"-e", "sfe",
			"-t", "hyb"],
		re.compile(r"95% confidence\n.*\n.*interval: \[ (.+), (.+)\]$", re.MULTILINE),
		re.compile(r"([0-9]+) samples", re.MULTILINE))
]


######################################
if TEST_RUN:
	TIMEOUT = 60
	REPETITIONS = 2
	model_instances = [
		("tandem_queue", "10"),
		("open_closed_queue", "30"),
		("queues_with_breakdowns", "80"),
		("oilpipes", "2-ray"),
		("database", "3-500"),
	]
	tools = [
		("modes.restart.es",
			[MODEST_BIN,
				"simulate",
				"models/{0}.{1}.jani",
				"-D",
				"--unsafe",
				"-J", "1",
				"-N", "2000000000",
				"--timeout", "{2}",
				"--rare", "Restart",
				"--levels", "ExpectedSuccess",
				"--max-run-length", "0"],
			re.compile(r"Confidence interval: *\[(.+), (.+)\]$", re.MULTILINE),
			re.compile(r"Runs used: *([0-9]+)$", re.MULTILINE)),
		("fig.restart.es",
			[FIG_BIN,
				"models/{0}.{1}.sa",
				"--acomp",
				"--stop-time", "{2}",
				"--timeout", "{2}",
				"-e", "restart",
				"-t", "es"],
			re.compile(r"95% confidence\n.*\n.*interval: \[ (.+), (.+)\]$", re.MULTILINE),
			re.compile(r"([0-9]+) samples", re.MULTILINE)),
	]
######################################


def run_case(instance, tool, csv_mean, csv_median):
	avg_count = 0
	err_count = 0
	avg_value = 0.0
	avg_runs = 0.0
	avg_time = 0.0
	avg_width = 0.0
	all_values = []
	all_runs = []
	all_times = []
	all_widths = []
	for rep in range(REPETITIONS):
		parameters = [p.format(*instance, TIMEOUT) for p in tool[1]]
		print("+ {0}".format(tool[0]), end = "", flush = True)
		start_time = timer()
		proc = subprocess.run(parameters, \
			                  stdout = subprocess.PIPE,\
			                  stderr = subprocess.STDOUT, \
			                  universal_newlines = True)
		time = timer() - start_time
		matches = tool[2].search(proc.stdout)
		if matches is None:
			err_count += 1
			value = "ERR"
			width = "n/a"
			print(": error ({0:.1f} s)".format(time))
		else:
			run_matches = tool[3].search(proc.stdout)
			cruns = int(run_matches.group(1))
			interval = (float(matches.group(1)), float(matches.group(2)))
			value = (interval[0] + interval[1]) / 2
			if cruns > 0 and value > 0.0:
				avg_count += 1
				width = interval[1] - interval[0]
				print(": {0} ± {1} ({3} runs, {2:.1f} s)". \
					format(value, width, time, cruns))
			else:
				err_count += 1
				value = "ERR"
				width = "n/a"
				print(": 0.0 ({2} runs, {1:.1f} s)". \
					format(value, time, cruns))
			if value != "ERR":
				avg_value += (value - avg_value) / avg_count
				avg_runs += (cruns - avg_runs) / avg_count
				avg_time += (time - avg_time) / avg_count
				avg_width += (width - avg_width) / avg_count
				all_values += [value]
				all_runs += [cruns]
				all_times += [time]
				all_widths += [width]
		file = open(RESULTS_FOLDER + "{0}.{1}.{2}.{3}.txt". \
			format(*instance, tool[0], rep), "w")
		cmd_parameters = ""
		for i in range(len(parameters)):
			if i != 0:
				cmd_parameters += " "
			cmd_parameters += "\"" + parameters[i] \
			                + "\"" if " " in parameters[i] else parameters[i]
		file.write("Command: {0}\n".format(cmd_parameters))
		file.write("Time: {0:.2f}\nValue: {1}\n".format(time, value))
		if matches is not None:
			file.write("Interval: [{0}, {1}]\n".format(interval[0], interval[1]))
			file.write("Width: {0}\n".format(width))
		file.write("\nTool output:\n")
		file.write(proc.stdout)
		file.close()
	# Compute median values
	means = [avg_value,avg_runs,avg_time,avg_width]
	medians = []
	for l in [all_values,all_runs,all_times,all_widths]:
		if (len(l)>0):
			l.sort()
			medians += [l[int(len(l)/2)]]
	# Dump to files
	for csv,l in zip([csv_mean,csv_median],[means,medians]):
		csv.write(";{0}".format(str(l[0]) if err_count < REPETITIONS else "ERR_VALUE"))
		csv.write(";{0}".format(str(l[1]) if err_count < REPETITIONS else "ERR_RUNS"))
		csv.write(";{0}".format(str(l[2]) if err_count < REPETITIONS else "ERR_TIME"))
		csv.write(";{0}".format(str(l[3]) if err_count < REPETITIONS else "ERR_ABSWIDTH"))
		csv.write(";{0}".format(str(l[3]/l[0]) if err_count < REPETITIONS else "ERR_RELWIDTH"))


def run(models, tools):
	# Create CSV header
	csv_mean = open("{0}/ALL_means.csv".format(RESULTS_FOLDER), "w")
	csv_median = open("{0}/ALL_medians.csv".format(RESULTS_FOLDER), "w")
	for csv in [csv_mean,csv_median]:
		csv.write("model;instance")
		for tool in tools:
			csv.write(";{0}-value".format(*tool))
			csv.write(";{0}-runs".format(*tool))
			csv.write(";{0}-time".format(*tool))
			csv.write(";{0}-absolute-width".format(*tool))
			csv.write(";{0}-relative-width".format(*tool))
		csv.write("\n")
	NCMDS = len(tools)*len(models)*REPETITIONS
	print("Experiments started at {0}".format(datetime.datetime.now()))
	print("Storing results in " + RESULTS_FOLDER)
	print("Running {0} commands (duration ~ {1})". \
		format(NCMDS, str(datetime.timedelta(seconds=NCMDS*(TIMEOUT+1)))))
	# Run experiments
	for instance in models:
		print("{0}.{1}".format(*instance))
		csv_mean.write("{0};{1}".format(*instance))
		csv_median.write("{0};{1}".format(*instance))
		for tool in tools:
			run_case(instance, tool, csv_mean=csv_mean, csv_median=csv_median)
		csv_mean.write("\n")
		csv_median.write("\n")
	csv_mean.close()
	csv_median.close()
	print("Experiments finished at {0}". \
			format(datetime.datetime.now()))


run(model_instances, tools)

