#!/usr/bin/env bash
#
# Author:  Carlos E. Budde
# Date:    06.10.2019
# License: GPLv3
#

## Configuration
#
MAXCPUS=1;
FIG_BIN="/home/tacas20ae/fig/fig";
MODELS_DIR="models/";
MODELS_EXT="sa";
#
PLOT_ONLY=0  # set to (only) plot already existing results
TEST_RUN=1   # set for small debug run
FLAT=0;      # run also non-RES experiment?
#
if [[ $PLOT_ONLY && $PLOT_ONLY -ne 0 ]]; then
	DIR="results_0";  # directory with results to plot (PLOT_ONLY)
else
#
if [[ $TEST_RUN && $TEST_RUN -ne 0 ]]; then
	TIMEOUT=35;
	NUM_ITERS=2;
	unset -v FLAT;
	MODELS=("2tandem_queue_M" "3tandem_queue_M")
else
	TIMEOUT=1200;  # in seconds
	NUM_ITERS=13;
	MODELS=(\
		"2tandem_queue_M"\
		"3tandem_queue_M"\
		"3tandem_queue_NM"\
	)
fi
#
fi # !PLOT_ONLY: line 21
#
AH_RUNS=(\
	# 2tandem_queue_M: 0.63*Queue1 + Queue2
	'-e restart --adhoc 63*q1+100*q2 --thresholds-adhoc 200:3,300:2,400:2,500:3,600:3,700:3,800:2,900:3,1000:3,1100:3,1200:3,1300:3,1400:2,1500:3,1600:3,1700:3,1800:3,1900:3,2000:3,2100:3,2200:3,2300:3,2400:3,2500:3,2600:3,2700:3,2800:3,2900:5 --parlare 1'\
	# 3tandem_queue_M: 0.37*Queue1 + 0.63*Queue2 + Queue3
	'-e restart --adhoc 37*q1+63*q2+100*q3 --thresholds-adhoc 200:2,300:2,400:2,500:2,600:2,700:3,800:2,900:3,1000:3,1100:3,1200:2,1300:3,1400:3,1500:3,1600:2,1700:3,1800:3,1900:3,2000:2,2100:3,2200:3,2300:3,2400:3,2500:3,2600:2,2700:3,2800:3,2900:3,3000:3,3100:2,3200:3,3300:3,3400:3 --parlare 1'\
	# 3tandem_queue_NM: 0.13*Queue1 + 0.5*Queue2 + Queue3
	'-e restart --adhoc 13*q1+50*q2+100*q3 --thresholds-adhoc 200:4,300:5,400:6,500:7,600:7,700:8,800:7,900:8,1000:8,1100:8,1200:9,1300:9,1400:9,1500:10,1600:9,1700:6 --parlare 1'\
)
# NOTE: changes in AUTO_RUNS must be reflected in
#       RUN_NAMES and gnuplot invocation below
AUTO_RUNS=(\
	"-e restart --amono      --parlare 1"\
	"-e restart --amono -g 4 --parlare 1"\
	"-e restart --amono -g 5 --parlare 1"\
	"-e restart --amono -g 7 --parlare 1"\
	"-e restart --amono -g 9 --parlare 1"\
)
#	"-e restart --amono -t es --parlare 1"\
RUN_NAMES=(\
	"AUTO_3"\
	"AUTO_4"\
	"AUTO_5"\
	"AUTO_7"\
	"AUTO_9"\
)
#	"AUTO_ES"\

if [[ !($PLOT_ONLY && $PLOT_ONLY -ne 0) ]]; then

## Test & set environment
if [ ! -f ${FIG_BIN} ]; then
	echo "[ERROR] fig binary file \"${FIG_BIN}\" not found";
	exit 1;
elif [ ! -x ${FIG_BIN} ]; then
	echo "[ERROR] fig binary file \"${FIG_BIN}\" is not executable";
	exit 1;
fi
echo "Experiments started at `date`";
#
#ITERS=$(printf " --stop-time $TIMEOUT%.0s" `seq $NUM_ITERS`);
ITERS="--stop-time ${TIMEOUT}s";
#
DIR="results";
N=0; while [ -e "${DIR}_${N}" ]; do ((N++)); done
DIR="${DIR}_${N}";
mkdir -p $DIR;
echo "Storing results in ${DIR}";
#
format_time() {
	# Print passed seconds as "hh:mm:ss", e.g. 3723 --> "1:02:03"
	local HOURS=$(($1 / 3600));
	local MINUTES=$((($1 % 3600) / 60));
	printf "%d:%02d:%02d" $HOURS $MINUTES $(($1 % 60));
}
poll_till_free() {
	# Poll until there are less jobs called 'fig' than CPUs
	# Can take an argument (part of the command) for thinner filtering
	if [ -z "$MAXCPUS" ]; then
		local MAXJOBS=`nproc --all`;
	else
		local MAXJOBS=$MAXCPUS;
	fi
	while  # do-while syntax sugar for bash:
		local RUNNING=`ps -fC "fig" | grep "\<$1" | wc -l`
		[ $RUNNING -ge $MAXJOBS ]
	do
		sleep 10s;
	done
	sleep 0.1;  # give previous job a little while to sink in
}

## Run experiments
NCMDS=$(((1+${#RUN_NAMES[@]})*${#MODELS[@]}*${NUM_ITERS}));
TOTAL_TIME=`format_time $(((${NCMDS}*${TIMEOUT}*2)/${MAXCPUS}))`;
echo "Running ${NCMDS} experiments (duration ~ ${TOTAL_TIME})";
#
for (( i=0 ; i<${#MODELS[@]} ; i++ )); do
	MODEL="${MODELS[i]}";
	MODEL_FILE="${MODELS_DIR}/${MODEL}.${MODELS_EXT}";
	RUN_FIG="$FIG_BIN $ITERS ${MODEL_FILE}";
	LOG="${DIR}/${MODEL}";
	echo "${MODEL}:";
	for (( k=0 ; k<${NUM_ITERS} ; k++ )); do
		echo -n " [${k}]";
		# Run with ad hoc ifun (there's one per model)
		poll_till_free fig;
		echo -n " AH";
		$RUN_FIG ${AH_RUNS[i]} 1>>${LOG}"_AH_${k}.out" \
		                       2>>${LOG}"_AH_${k}.err" &
		# Run with automatic ifuns
		for (( j=0 ; j<${#AUTO_RUNS[@]} ; j++ )); do
			poll_till_free fig;
			NAME=${RUN_NAMES[j]}
			echo -n " ${NAME}";
			$RUN_FIG ${AUTO_RUNS[j]} 1>>${LOG}"_${NAME}_${k}.out" \
			                         2>>${LOG}"_${NAME}_${k}.err" &
		done
		# If requested: Run with classical Monte Carlo (not RES)
		if [[ $FLAT && $FLAT -ne 0 ]]; then
			poll_till_free fig;
			echo -n " FLAT";
			$RUN_FIG --flat --parlare 1 1>>${LOG}"_FLAT_${k}.out" \
			                            2>>${LOG}"_FLAT_${k}.err" &
		fi
		echo;
	done
done
echo "Waiting for all experiments to finish";
wait;

## Summarise data
echo "Summarising experimental results";
ALL_RUNS=("AH" ${RUN_NAMES[@]});
if [[ $FLAT && $FLAT -ne 0 ]]; then ALL_RUNS=(${ALL_RUNS[@]} "FLAT"); fi
for MODEL in "${MODELS[@]}"; do
	for RUN in "${ALL_RUNS[@]}"; do
		grep "Computed estimate:" --after=14 ${DIR}/${MODEL}_${RUN}_*.out \
			&> ${DIR}/${MODEL}_${RUN}_ALL.txt;
	done
done
SUMMARY_ESTIM="${DIR}/summary_estimate.txt";
SUMMARY_PRECI="${DIR}/summary_precision.txt";
## (Point) Estimates
for RUN in ${ALL_RUNS[@]}; do
	echo -e "\n\n${RUN}\n";
	for MODEL in ${MODELS[@]}; do
		echo -e "Model: ${MODEL}\n";
		grep "Computed estimate:" "${DIR}/${MODEL}_${RUN}_ALL.txt" \
			| cut -f 7 -d" ";
		echo;
	done;
done &> "${SUMMARY_ESTIM}";
## CI precisions
for RUN in ${ALL_RUNS[@]}; do
	echo -e "\n\n${RUN}\n";
	for MODEL in ${MODELS[@]}; do
		echo -e "Model: ${MODEL}\n";
		grep "95% confidence" --after=1 "${DIR}/${MODEL}_${RUN}_ALL.txt" \
			| grep precision \
			| cut -f 2 -d":";
		echo;
	done;
done &> "${SUMMARY_PRECI}";
echo "Summaries dumped in ${SUMMARY_ESTIM} and ${SUMMARY_PRECI}";

set -e;  # Needed for error checks below

## Create plot data from summary
PYSUMMARY="summary_for_plot.py";
command -v python3 &>/dev/null || \
	(echo "[ERROR] python3 not found in the environment" && exit 1);
[[ -f "${PYSUMMARY}" ]] || \
	(echo "[ERROR] Script ${PYSUMMARY} not found" && exit 1);
python3 "${PYSUMMARY}" "${SUMMARY_PRECI}";

fi # !PLOT_ONLY: line 67

## Plot
echo "Plotting results";
PLOTSCRIPT="plot_precision.gp";
command -v gnuplot &>/dev/null || \
	(echo "[ERROR] gnuplot not found in the environment" && exit 1);
[[ -f "${PLOTSCRIPT}" ]] || \
	(echo "[ERROR] Script ${PLOTSCRIPT} not found" && exit 1);
gnuplot -e "
	DIR='${DIR}';
	AH='AH';
	AUTO_0='${RUN_NAMES[0]}';
	AUTO_1='${RUN_NAMES[1]}';
	AUTO_2='${RUN_NAMES[2]}';
	AUTO_3='${RUN_NAMES[3]}';
	AUTO_4='${RUN_NAMES[4]}';
	" "${PLOTSCRIPT}";

echo "Experiments finished at `date`";

exit 0;

